# -*- mode: python; coding: utf-8; -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""
The :py:mod:`aiatools.attributes` module provides :py:class:`~aiatools.algebra.Functor` for querying App Inventor
projects.

.. testsetup::

    from aiatools.aia import AIAFile
    from aiatools.attributes import *
    from aiatools.block_types import *
    from aiatools.component_types import *
    project = AIAFile('test_aias/LondonCholeraMap.aia')
"""

from aiatools.algebra import ComputedAttribute
from .algebra import Functor, NotExpression
from .common import Block, BlockKind, Component
from .selectors import select
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable


__author__ = 'Evan W. Patton <ewpatton@mit.edu>'


class NamedAttribute(Functor):
    """
    The :py:class:`NamedAttribute` is a :py:class:`Functor` that retrieves the value of a specific field on entities in
    an App Inventor project.

    Example
    _______
    Retrieve any Form entities in the project:

        >>> project.components(NamedAttribute('type') == Form)
        [Screen('Screen1')]

    Parameters
    __________
    name : basestring
        The name of the attribute to be retrieved.
    """

    # noinspection PyShadowingNames
    def __init__(self, name):
        self.name = name

    def __call__(self, obj, *args, **kwargs):
        if hasattr(obj, self.name):
            return getattr(obj, self.name)
        return None

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, NamedAttribute):
            return self.name == other.name
        else:
            return super(NamedAttribute, self).__eq__(other)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, repr(self.name))


class NamedAttributeTuple(Functor):
    """
    The :py:class:`NamedAttributeTuple` is a :py:class:`Functor` that retrieves the value of a field on entities in an
    App Inventor project. Unlike :py:class:`NamedAttribute`, it can take more than one name as a tuple. The given field
    names are searched in order until one is found. This is useful for presenting a single :py:class:`Functor` over
    synonyms, for example, ``'name'`` gives the name of a :py:class:`Component` whereas ``'component_name'`` gives the
    name of a block's component (if any).

    Parameters
    ----------
    names : (str, unicode)
        The name(s) of the attribute(s) to be retrieved. Giving a 1-tuple is less efficient than defining and using the
        equivalent :py:class:`NamedAttribute` instance.
    """
    def __init__(self, names):
        super(NamedAttributeTuple, self).__init__()
        self.names = names

    def __call__(self, obj, *args, **kwargs):
        for _name in self.names:
            if hasattr(obj, _name):
                return getattr(obj, _name)
        return None

    def __hash__(self):
        return hash(self.names)

    def __eq__(self, other):
        if isinstance(other, NamedAttributeTuple):
            return self.names == other.names
        else:
            return super(NamedAttributeTuple, self).__eq__(other)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.names)


def has_ancestor(target=None):
    """
    Constructs a new ComputedAttribute that accepts an entity if and only if the entity has an ancestor that matches the
    given ``target`` clause.

    Example
    -------
    Count the number of :py:data:`~.block_type.text` blocks with an ancestor that is a
    :py:data:`~.block_type.logic_compare` block.

        >>> project.blocks((type == text) & has_ancestor(type == logic_compare)).count()
        1

    Parameters
    ----------
    target : Expression
        An :py:class:`Expression` to use for testing ancestors.

    Returns
    -------
    ComputedAttribute
        A new ComputedAttribute that will walk the entity graph using the ``parent`` field and test whether any
        ``parent`` matches ``target``.
    """
    def checkAncestor(b):
        if b is None:
            return False
        b = b.parent  # Skip b
        while b is not None:
            if target is None:
                return True
            elif isinstance(target, Callable) and target(b):
                return True
            elif b is target:
                return True
            b = b.parent
        return False
    return ComputedAttribute(checkAncestor)


def has_descendant(target=None):
    """
    Constructs a new ComputedAttribute that accepts an entity if and only if the entity has a descendant that matches
    the given ``target`` clause.

    Example
    -------
    Count the number of top-level blocks that have control_if blocks as descendants.

        >>> project.blocks(top_level & has_descendant(type == controls_if)).count()
        1

    Parameters
    ----------
    target : Expression
        An :py:class:`Expression` to use for testing descendants.

    Returns
    -------
    ComputedAttribute
        A new ComputedAttribute that will walk the entity graph using the ``children`` field and test whether any
        descendant in the subgraph matches ``target``.
    """
    def checkDescendant(b):
        if b is None:
            return False
        if not hasattr(b, 'children'):
            return False
        for child in b.children():
            if target is None:
                return True
            elif isinstance(target, Callable) and target(child):
                return True
            elif child is target:
                return True
            elif checkDescendant(child):
                return True
        return False
    return ComputedAttribute(checkDescendant)


def _root_block(block):
    """
    Looks up the root of the stack of blocks containing the given ``block``.

    Example
    -------

    Parameters
    ----------
    block : Block
        The block of interest for which the root block will be obtained.

    Returns
    -------
    Block
        The block at the root of the block stack containing ``block``.
    """
    if not block:
        return block
    while block.logical_parent:
        block = block.logical_parent
    return block


root_block = ComputedAttribute(_root_block)


class HeightAttribute(Functor):
    """
    :py:class:`HeightAttribute` class is used to memoize the heights of the forest representing an App Inventor
    project. Use :py:data:`aiatools.attributes.height` to benefit from the memoization feature.

    Example
    -------
    Get the heights of the block stacks in the project.

        >>> project.blocks(top_level).map(height)
        [2, 6]
    """
    def __init__(self):
        self.precomputed = {}

    # noinspection PyShadowingNames
    def __call__(self, *args, **kwargs):
        block_or_component = args[0]
        if block_or_component in self.precomputed:
            return self.precomputed[block_or_component]
        else:
            try:
                height = max(self(x) for x in block_or_component.children()) + 1
            except ValueError:
                height = 0
            self.precomputed[block_or_component] = height
            return height


class DepthAttribute(Functor):
    """
    :py:class:`DepthAttribute` class is used to memoize the depths of entities in the forest representing an App
    Inventor project. Use :py:data:`aiatools.attributes.depth` to benefit from the memoization feature.

    Example
    -------
    Get the depth of all leaves in the project.

        >>> project.blocks(leaf).map(depth)
        [2, 2, 4, 4, 4, 6, 5, 5, 5, 5, 5]
    """
    def __init__(self):
        self.precomputed = {}

    @staticmethod
    def _get_parent(block_or_component):
        if hasattr(block_or_component, 'logical_parent'):
            return block_or_component.logical_parent
        else:
            return block_or_component.parent

    # noinspection PyShadowingNames
    def __call__(self, *args, **kwargs):
        block_or_component = DepthAttribute._get_parent(args[0])
        depth = 0
        while block_or_component is not None:
            depth += 1
            block_or_component = DepthAttribute._get_parent(block_or_component)
        return depth


class _MutationHelper(Functor):
    """
    :py:class:`_MutatorHelper` is a helper class used for generating new :py:class:`Functor` for retrieving mutation
    fields on a block. :py:class:`_MutationHelper` interns the instances so that the returned items are the same
    instance, that is:

        >>> mutation.component_type is mutation.component_type
        True

    :py:class:`_MutationHelper` is accessed through the :py:data:`mutation` instance.
    """
    _interned = {}

    def __init__(self, child=None):
        self.child = child

    def __call__(self, *args, **kwargs):
        if hasattr(args[0], 'mutation'):
            _mutation = args[0].mutation
            if _mutation is not None:
                if self.child is None:
                    return _mutation
                elif self.child in _mutation:
                    return _mutation[self.child]
        return None

    def __getattr__(self, item):
        if item not in _MutationHelper._interned:
            _MutationHelper._interned[item] = _MutationHelper(item)
        return _MutationHelper._interned[item]


class _FieldHelper(Functor):
    """
    :py:class:`_FieldHelper` is a helper class used for generating new :py:class:`Functor` for retrieving fields of
    blocks. :py:class:`_FieldHelper` interns the instances so that the returned items are the same instace, that is:

        >>> fields.OP is fields.OP
        True

    :py:class:`_FieldHelper` is accessed through the :py:data:`field` instance.
    """
    _interned = {}

    def __init__(self, child=None):
        self.child = child

    def __call__(self, *args, **kwargs):
        if hasattr(args[0], 'fields'):
            _fields = args[0].fields
            if _fields is not None:
                if self.child is None:
                    return _fields
                elif self.child in _fields:
                    return _fields[self.child]
        return None

    def __getattr__(self, item):
        if item not in _FieldHelper._interned:
            _FieldHelper._interned[item] = _FieldHelper(item)
        return _FieldHelper._interned[item]


type = NamedAttributeTuple(('type', 'component_type'))
"""Returns the type of the entity."""

name = NamedAttributeTuple(('name', 'instance_name'))
"""Returns the name of the entity."""


def _kind(block):
    try:
        return block.kind
    except AttributeError:
        return None


kind = NamedAttribute("kind")
"""Returns the kind of the entity."""


def _compute_external(x):
    if isinstance(x, Component):
        return x.type.external
    elif isinstance(x, Block):
        if x.has_mutation('component_type'):
            comp_typename = x.mutation['component_type']
            comp_type = x.screen.project.extensions[comp_typename]
            return comp_type.external if comp_type else False
    else:
        return False  # Cannot determine externality of unknown object


external = ComputedAttribute(_compute_external)
"""Returns true if the component is an extension."""

version = NamedAttribute('version')
"""Returns the version number for the entity."""

category = NamedAttributeTuple(('category', 'category_string'))
"""Returns the category for the entity."""

help_string = NamedAttribute('help_string')
"""Returns the help string for the entity."""

show_on_palette = NamedAttribute('show_on_palette')
"""Returns true if the entity is shown in the palette."""

visible = NamedAttribute('visible')
"""Returns true if the entity is visible."""

non_visible = NotExpression(visible)
"""Returns true if the entity is nonvisible."""

icon_name = NamedAttribute('iconName')
"""Returns the icon for the component."""

return_type = NamedAttribute('return_type')
"""Gets the return type of the block."""

generic = NamedAttribute('generic')
"""Tests whether the block is a generic component block."""

disabled = NamedAttribute('disabled')
"""Tests whether the entity is disabled."""

logically_disabled = NamedAttribute('logically_disabled')
"""Tests whether the block is logically disabled, either because it is explicitly disabled or is contained within a
disabled subtree."""

logically_enabled = ~logically_disabled
"""Tests whether the block is logically enabled."""

enabled = NamedAttribute('Enabled') | ~disabled
"""Tests whether the entity is enabled."""

top_level = ComputedAttribute(lambda b: isinstance(b, Block) and b.parent is None)
"""Tests whether the block is at the top level."""

parent = NamedAttribute('parent')
"""Gets the parent of the entity."""

mutation = _MutationHelper()
"""
Tests whether a block has a mutation specified. One can also use :py:data:`mutation` to obtain accessors for specific
mutation fields, for example:

.. doctest::

    >>> project.blocks(mutation.component_type == Button)
    [...]

will retrieve all blocks that have a mutation where the component_type key is the Button type.
"""

fields = _FieldHelper()
"""
:py:data:`fields` is a generator for :py:class:`Functor` to retrieve the values of fields in a block. For example:

.. doctest::

    >>> project.blocks(logic_compare).map(fields.OP)
    ['EQ']
"""

depth = DepthAttribute()
"""
Computes the depth of the entity with its tree. For components, this will be the number of containers from the Screen.
For blocks, this will be the number of logical ancestors to the top-most block of the block stack.

.. doctest::

    >>> project.components(type == Marker).avg(depth)
    2.0
"""

height = HeightAttribute()
"""
Computes the height of the tree from the given entity. This will be the longest path from the node to one of its
children. For leaf nodes, the height is 0.

.. doctest::

    >>> project.blocks(top_level).avg(height)
    4.0
"""

is_procedure = (type == 'procedures_defreturn') | (type == 'procedures_defnoreturn')
"""
Returns True if the type of a block is a procedure definition block, either :py:data:`procedures_defreturn` or
:py:data:`procedures_defnoreturn`

.. doctest::

    >>> with AIAFile('test_aias/ProcedureTest.aia') as proc_project:
    ...    proc_project.blocks(is_procedure).count()
    7
"""

is_called = ComputedAttribute(lambda x: not select(x).callers().empty())
"""
Returns True if the entity in question is called by some other block in the code.

.. todo::

    Add a mechanism for describing the call graph internal to components so that, for example,
    ``project.blocks(mutation.event_name == 'GotText').callers()`` should be non-empty for any ``Get`` method call
    blocks in the screen for the same ``instance_name``.
"""

leaf = ComputedAttribute(lambda x: len(x.children()) == 0)
"""
Returns True if the entity is a leaf in the tree (i.e., it has no children).

.. doctest::

    >>> project.blocks(leaf).count(group_by=type)
    {'text': 3, 'lexical_variable_get': 6, 'color_blue': 2}
"""

declaration = (type == 'component_event') | (type == 'global_declaration') | is_procedure
"""
"""

statement = kind == BlockKind.STATEMENT
"""
"""

value = kind == BlockKind.VALUE

"""
.. testcleanup::

    project.close()
"""



import unittest
from aiatools.algebra import *


class AndTest(unittest.TestCase):
    def test_eval_true(self):
        x = and_(identity, True)
        self.assertTrue(x(True))

    def test_eval_false(self):
        x = and_(identity, True)
        self.assertFalse(x(False))

    def test_repr(self):
        x = and_(True, True)
        self.assertEqual('True & True', '%r' % x)


class OrTest(unittest.TestCase):
    def test_eval_true(self):
        x = or_(identity, False)
        self.assertTrue(x(True))

    def test_eval_false(self):
        x = or_(identity, False)
        self.assertFalse(x(False))

    def test_repr(self):
        x = or_(True, True)
        self.assertEqual('True | True', '%r' % x)


class NotTest(unittest.TestCase):
    def test_eval_true(self):
        x = not_(identity)
        self.assertTrue(x(False))

    def test_eval_false(self):
        x = not_(identity)
        self.assertFalse(x(True))

    def test_repr(self):
        x = not_(True)
        self.assertEqual('~True', '%r' % x)
        x = not_(not_(True) == False)
        self.assertEqual('~(~True == False)', '%r' % x)

    def test_invert(self):
        x = not_(True)
        self.assertTrue(~x)



# -*- mode: python; coding: utf-8; -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""

"""

from .common import BlockCategory, BlockKind, BlockType

DECL = BlockKind.DECLARATION
STMT = BlockKind.STATEMENT
VAL = BlockKind.VALUE
MUT = BlockKind.MUTATION

__author__ = 'Evan W. Patton <ewpatton@mit.edu>'


def define_block_type(name, category, kind):
    globals()[name] = BlockType(name, category, kind)


Control = BlockCategory('Control')
Logic = BlockCategory('Logic')
Math = BlockCategory('Math')
Text = BlockCategory('Text')
Lists = BlockCategory('Lists')
Colors = BlockCategory('Colors')
Variables = BlockCategory('Variables')
Procedures = BlockCategory('Procedures')
Dictionaries = BlockCategory('Dictionaries')
# Components is plural to prevent collision with aiatools.common.Component
Components = BlockCategory('Components')

# Control category blocks
for _name, _kind in [
        ('controls_if', STMT),
        ('controls_forRange', STMT),
        ('controls_forEach', STMT),
        ('controls_while', STMT),
        ('controls_choose', VAL),
        ('controls_do_then_return', VAL),
        ('controls_eval_but_ignore', STMT),
        ('controls_openAnotherScreen', STMT),
        ('controls_openAnotherScreenWithStartValue', STMT),
        ('controls_getStartValue', VAL),
        ('controls_closeScreen', STMT),
        ('controls_closeScreenWithValue', STMT),
        ('controls_closeApplication', STMT),
        ('controls_getPlainStartText', VAL),
        ('controls_closeScreenWithPlainText', STMT),
        ('controls_break', STMT)]:
    define_block_type(_name, Control, _kind)

# Logic category blocks
for _name in ['logic_boolean', 'logic_false', 'logic_negate', 'logic_compare', 'logic_operation', 'logic_or']:
    define_block_type(_name, Logic, VAL)

# Math category blocks
for _name in ['math_number', 'math_compare', 'math_add', 'math_subtract', 'math_multiply', 'math_division',
              'math_power', 'math_bitwise', 'math_random_int', 'math_random_float', 'math_random_set_seed',
              'math_on_list', 'math_single', 'math_abs', 'math_neg', 'math_round', 'math_ceiling', 'math_floor',
              'math_divide', 'math_trig', 'math_cos', 'math_tan', 'math_atan2', 'math_convert_angles',
              'math_format_as_decimal', 'math_is_a_number', 'math_convert_number']:
    define_block_type(_name, Math, STMT if _name == 'math_random_set_seed' else VAL)

# Text category blocks
for _name in ['text', 'text_join', 'text_length', 'text_isEmpty', 'text_compare', 'text_trim', 'text_changeCase',
              'text_starts_at', 'text_contains', 'text_split', 'text_split_at_spaces', 'text_segment',
              'text_replace_all', 'obfuscated_text', 'text_is_string']:
    define_block_type(_name, Text, VAL)

# renamed block
obsfucated_text = globals()['obfuscated_text']

# Lists category blocks
for _name, _kind in [
        ('lists_create_with', VAL),
        ('lists_create_with_item', VAL),
        ('lists_add_items', STMT),
        ('lists_is_in', VAL),
        ('lists_length', VAL),
        ('lists_is_empty', VAL),
        ('lists_pick_random_item', VAL),
        ('lists_position_in', VAL),
        ('lists_select_item', VAL),
        ('lists_insert_item', STMT),
        ('lists_replace_item', STMT),
        ('lists_remove_item', STMT),
        ('lists_append_list', STMT),
        ('lists_copy', VAL),
        ('lists_is_list', VAL),
        ('lists_to_csv_row', VAL),
        ('lists_to_csv_table', VAL),
        ('lists_from_csv_row', VAL),
        ('lists_from_csv_table', VAL),
        ('lists_lookup_in_pairs', VAL),
        ('lists_join_with_separator', VAL)]:
    define_block_type(_name, Lists, _kind)

# Dictionaries category blocks
for _name, _kind in [
        ('dictionaries_create_with', VAL),
        ('pair', VAL),
        ('dictionaries_lookup', VAL),
        ('dictionaries_set_pair', STMT),
        ('dictionaries_delete_pair', STMT),
        ('dictionaries_recursive_lookup', VAL),
        ('dictionaries_recursive_set', STMT),
        ('dictionaries_getters', VAL),
        ('dictionaries_get_values', VAL),
        ('dictionaries_is_key_in', VAL),
        ('dictionaries_length', VAL),
        ('dictionaries_alist_to_dict', VAL),
        ('dictionaries_dict_to_alist', VAL),
        ('dictionaries_copy', VAL),
        ('dictionaries_combine_dicts', STMT),
        ('dictionaries_walk_tree', VAL),
        ('dictionaries_walk_all', VAL),
        ('dictionaries_is_dict', VAL)]:
    define_block_type(_name, Dictionaries, _kind)

# Colors category blocks
for _name in ['color_black', 'color_white', 'color_red', 'color_pink', 'color_orange', 'color_yellow', 'color_green',
              'color_cyan', 'color_blue', 'color_magenta', 'color_light_gray', 'color_gray', 'color_dark_gray',
              'color_make_color', 'color_split_color']:
    define_block_type(_name, Colors, VAL)

# Variables category blocks
for _name, _val in [
        ('global_declaration', DECL),
        ('lexical_variable_get', VAL),
        ('lexical_variable_set', STMT),
        ('local_declaration_statement', STMT),
        ('local_declaration_expression', VAL)]:
    define_block_type(_name, Variables, _val)

# Procedures category blocks
for _name in ['procedures_defnoreturn', 'procedures_defreturn', 'procedures_callnoreturn', 'procedures_callreturn']:
    define_block_type(_name, Procedures, VAL if 'call' in _name else DECL)

# Component category blocks
for _name, _kind in [('component_event', DECL), ('component_method', MUT), ('component_set_get', MUT),
                     ('component_component_block', VAL)]:
    define_block_type(_name, Components, _kind)



# -*- mode: python; coding: utf-8; -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""
The :py:mod:`aiatools.selectors` modules provides high level selection and aggregation operations that can be executed
over a project.

.. testsetup:
    from aiatools import *
    project = AIAFile('test_aias/LondonCholeraMap.aia')
"""


from aiatools.algebra import Expression, AndExpression, identity
from aiatools.common import Block, Component, ComponentType, FilterableDict, RecursiveIterator
from aiatools.block_types import procedures_defnoreturn, procedures_defreturn, procedures_callnoreturn, \
    procedures_callreturn
from functools import reduce

__author__ = 'Evan W. Patton <ewpatton@mit.edu>'


class AggregateOperations:
    """
    :py:class:`AggregateOptions` is a mixin class that provides functionality for aggregation operations on selections.
    """

    # noinspection PyTypeChecker
    def count(self, group_by=None):
        """
        Counts the number of entities in a collection, optionally grouping by the ``group_by`` argument. ``group_by``
        will be called each element in the collection.

        .. doctest::

            >>> project.components().count()
            4
            >>> project.blocks(top_level).count(group_by=type)
            {'component_event': 2}
            >>> project.blocks().count(group_by=(type, mutation.component_type))
            {('component_event', 'Button'): 1, ('component_method', 'Map'): 2, ('component_event', 'Map'): 1, ('component_set_get', 'Marker'): 4}

        :param group_by: If given, a :py:class:`~aiatools.algebra.Functor` or tuple thereof, the value(s) of which
            will be used to group the entities in the collection for counting.
        :type group_by: tuple[aiatools.algebra.Functor or callable] or aiatools.algebra.Functor or callable or None
        :return: The count of the number of entities or a dictionary mapping unique ``group_by`` values to counts.
        :rtype: int or dict[(Atom, str) or Atom or str, int]
        """
        if group_by is None:
            result = 0
            for _ in self:
                result += 1
            return result
        elif not isinstance(group_by, tuple):
            group_by = (group_by,)
        result = FilterableDict()
        for value in self:
            attr = tuple(x(value) for x in group_by)
            if None not in attr:
                if len(attr) == 1:
                    attr = attr[0]
                if attr in result:
                    result[attr] += 1
                else:
                    result[attr] = 1
        return result

    def avg(self, func, group_by=None):
        """
        Averages the values of ``func`` applied to the collection, optionally grouping by the ``group_by`` argument.
        ``group_by`` will be called on each element in the collection.

        .. doctest::

            >>> project.blocks(top_level).avg(height)
            4.0
            >>> project.blocks().avg(depth)
            3.347826086956522

        :param func: The function to apply to the entities in the collection for which the average will be computed.
        :type func: aiatools.algebra.Functor or callable
        :param group_by: If given, a :py:class`~aiatools.algebra.Functor` or tuple thereof, the value(s) of which will
            be used to group the entities in the collection for averaging.
        :type group_by: tuple[aiatools.algebra.Functor or callable] or aiatools.algebra.Functor or callable or None
        :return: The average of the values of ``func`` applied to the entities of the collection or a dictionary
            mapping the value(s) of ``group_by`` applied to the entities to the average of ``func`` applied to the
            entities in the subset identified by the dictionary key.
        :rtype: int or dict[(Atom, str) or Atom or str, int]
        """
        if group_by is None:
            results = list(map(func, self))
            return sum(results) / len(results)
        elif not isinstance(group_by, tuple):
            group_by = (group_by,)
        state = {}
        for value in self:
            attr = tuple(x(value) for x in group_by)
            if None not in attr:
                if len(attr) == 1:
                    attr = attr[0]
                if attr in state:
                    state[attr]['sum'] += func(value)
                    state[attr]['count'] += 1
                else:
                    state[attr] = {'sum': func(value), 'count': 1}
        return FilterableDict({k: v['sum'] / v['count'] for k, v in state})

    def max(self, func, group_by=None):
        """
        Obtains the maximum value of ``func`` applied to the collection, optionally grouping by the ``group_by``
        argument. ``group_by`` will be called on each element in the collection.

        .. doctest::

            >>> project.blocks(type == component_event).max(height)
            6

        :param func: The function to apply to the entities in the collection for which the maximum will be computed.
        :type func: aiatools.algebra.Functor or callable
        :param group_by: If given, a :py:class:`~aiatools.algebra.Functor` or tuple thereof, the value(s) of which will
            be used to group the entities in the collection for determining the maximum.
        :type group_by:
        :return: The maximum of the values of ``func`` applied to the entities of the collection or a dictionary
            mapping the value(s) of ``group_by`` applied to the entities to the maximum of ``func`` applied to the
            entities in the subset identified by the dictionary key.
        :rtype: int or dict[(Atom, str) or Atom or str, int]
        """
        if group_by is None:
            return max(list(map(func, self)))
        elif not isinstance(group_by, tuple):
            group_by = (group_by,)
        result = FilterableDict()
        for value in self:
            attr = tuple(x(value) for x in group_by)
            if None not in attr:
                if len(attr) == 1:
                    attr = attr[0]
                attr_value = func(value)
                if (attr in result and result[attr] < attr_value) or attr not in result:
                    result[attr] = attr_value
        return result

    def min(self, func, group_by=None):
        """
        Obtains the minimum value of ``func`` applied to the collection, optionally grouping by the ``group_by``
        argument. ``group_by`` will be called on each element in the collection.

        .. doctest::

            >>> project.blocks(type == component_event).min(height)
            2
            >>> project.blocks(category == Components).min(height, group_by=type)
            {'component_event': 2, 'component_method': 1, 'component_set_get': 1}
            >>> project.blocks().min(height, group_by=(type, mutation.component_type))
            {('component_event', 'Button'): 2, ('component_method', 'Map'): 1, ('component_event', 'Map'): 6, ('component_set_get', 'Marker'): 1}

        :param func: The function to apply to the enetities in the collection for which the minimum will be computed.
        :type func: aiatools.algebra.Functor or callable
        :param group_by: If given, a :py:class:`~aiatools.algebra.Functor` or tuple thereof, the value(s) of which will
            be used to group the entities in the collection determining the minimum.
        :type group_by: tuple[aiatools.
        :return: The minimum of the values of ``func`` applied to the entities of the collection or a dictionary
            mapping the value(s) of ``group_by`` applied to the entities to the minimum of ``func`` applied to the
            entities in the subset identified by the dictionary key.
        :rtype: int or dict[(Atom, str) or Atom or str, int]
        """
        if group_by is None:
            return min(list(map(func, self)))
        elif not isinstance(group_by, tuple):
            group_by = (group_by,)
        result = FilterableDict()
        for value in self:
            attr = tuple(x(value) for x in group_by)
            if None not in attr:
                if len(attr) == 1:
                    attr = attr[0]
                attr_value = func(value)
                if (attr in result and result[attr] > attr_value) or attr not in result:
                    result[attr] = attr_value
        return result

    def empty(self):
        """
        Tests whether the result of the selector chains up to and including the current selection is empty.

        .. doctest::

            >>> project.components(type == Voting).empty()
            True
            >>> project.components(type == Button).empty()
            False

        :return: ``True`` if the selection is empty, otherwise ``False``.
        :rtype: bool
        """
        for _ in self:
            return False
        return True

    def __iter__(self):
        raise NotImplemented()


class Selectors:
    """
    :py:class:`Selectors` is a mixin class for collections that enables selecting entities related to entities in the
    collection.
    """
    def __init__(self, *args):
        pass

    def screens(self, test=None, *args):
        """
        Select the screens containing the elements in the collection.

        Returns
        -------
        Selector[Screen]
            A selector over the screens containing the entities in the collection.

        Todo
        ----
            - (ewpatton) Implement subset selection on screens
        """
        test = identity if test is None else test
        screens = {item.id: item for item in self
                   if ((isinstance(item.type, ComponentType) and item.type.name == 'Form') or item.type == 'Form')
                   and test(item)}
        block_screens = {item.screen.id: item.screen for item in self if (isinstance(item, Block) and test(item))}
        screens.update(block_screens)
        return Selector(screens)

    def components(self, test=None, *args):
        """
        Select the subset of entities in the current selection that are :py:class:`Component`.

        Returns
        -------
        Selector[Component]
            A selector for further selection of entities. The selector will only contain :py:class:`Component`.

        Todo
        ----
            - (ewpatton) Implement subset selection on components
        """
        test = identity if test is None else test
        return Selector({item.id: item for item in self if (isinstance(item, Component) and test(item))})

    def blocks(self, test=None, *args):
        """
        Select blocks under the current node. The following filters can be applied:

        Returns
        -------
        Selector[Block]
            A selector for further selection of entities. The selector will only contain :py:class:`Block`.

        Todo
        ----
            - (ewpatton) Implement subset selection on blocks.
        """
        test = identity if test is None else test
        return Selector({item.id: item for item in self if (isinstance(item, Block) and test(item))})

    def callers(self, *args):
        """
        Select any blocks that result in a call to any callable blocks in the selection.

        Note
        ----
            Only operates on procedures at this time. See TODO about adding support for components.

        Returns
        -------
        Selector[Block]
            A selector over the callers (if any) of the callable blocks contained in the collection. The selector will
            only contain :py:class:`Block`.

        Todo
        ----
            - (ewpatton) Implement subset selection on the blocks
            - (ewpatton) Add call graph so that component method/event blocks can be included
        """
        _filter = args[0] if len(args) > 0 else lambda x: True
        return Selector({block.id: block for item in self
                         if item.type in (procedures_defreturn(), procedures_defnoreturn())
                         for block in item.screen.blocks
                         if block.type in (procedures_callnoreturn(), procedures_callreturn()) and
                         block.fields['PROCNAME'] == item.fields['NAME'] and _filter(block)})

    def callees(self, *args, **kwargs):
        """
        Select procedure definition blocks (if any) that are called by the procedure call blocks in the collection.

        Note
        ----
            Only operates on procedures at this time. See TODO about adding support for components.

        Returns
        -------
        Selector[Block]
            A selector over the callees (if any) of the caller blocks contained in the collection. The selector will
            only container :py:class:`Block`.

        Todo
        ----
            - (ewpatton) Implement subset selection on the blocks
            - (ewpatton) Add call graph so that component method/event blocks can be included
        """
        pass

    def branch(self, branch_id):
        """
        Retrieve

        Parameters
        ----------
        branch_id : int
            Retrieve the __branch_id__th branch of any blocks in this collection that have a statement input. This can
            be used to walk an if-elseif-else block, for example.

        Todo
        ----
            - (ewpatton) Implementation
        """
        pass

    def map(self, functor):
        """
        Applies ``functor`` to the entities in the selection.

        Parameters
        ----------
        functor : callable

        Returns
        -------
        list
            A list in the value space of ``functor``.
        """
        return [functor(x) for x in self if functor(x) is not None]

    def select(self, selector):
        """
        Selects a subset of the entities in the selection for which ``selector(item)`` is True.

        Parameters
        ----------
        selector : Expression
            The expression to be applied to filter the current selection.

        Returns
        -------
        Selector[T]
            The subset of the selection. The exact contents of the subset depends on the type of content of the
            current selection.
        """
        return Selector({item.id: selector(item) for item in self if selector(item) is not None})

    def descendants(self, test=None, order='natural', skip_failures=False):
        """
        Selects all of the descendants of the entities in the current selection.

        Parameters
        ----------
        test : callable
            An optional test used to filter out items from the iteration. Default: None

        order : str
            The order of iteration. Options are 'natural', 'breadth', or 'depth'. Default: 'natural'

        skip_failures : bool
            If skip_failures is true and test is provided but fails for an element, the subtree starting at the element
            is pruned.

        Returns
        -------
        Selector[T]
            The descendants of the entities in the current selection, if any.
        """
        def order_type(m):
            return order if order != 'natural' else ('depth' if isinstance(m, Block) else 'breadth')
        return Selector({obj.id: obj if not test or test(obj) else None for match in self
                         for obj in RecursiveIterator(match, order_type(match), test, skip_failures)})

    def __iter__(self):
        raise NotImplemented()


class Selector(AggregateOperations, Selectors):
    """
    :py:class:`Selector` provides a lazily computed application of an expression over a collection. Items in the
    underlying collection can be accessed in three ways:

    1. Through use of an iterable. :py:class:`Selector` is an iterable, and iterating over it yields its values. There
       is also :py:meth:`iteritems`, which iterates over key, value pairs in the collection similar to how the
       iteritems method in :py:obj:`dict` works.
    2. Through an identifier/key. For example, accessing the Component called Button1 can be done with
       ``selection['Button1']``.
    3. Through an index. For example, if ``children`` is a :py:class:`Selector, ``children[5]`` will give the 6th
       element in the selection. This is not random access, but linear in ``n``, so if accessing elements in sequence is
       required using the iteration method is recommended.

    Selectors can also be called, in which case they will return a new Selector whose elements are given by applying
    the first argument of the function call to each element in the underlying collection.

    Parameters
    ----------
    collection : collections.Iterable[{id}] or dict[{id}]
        A collection of objects
    """
    def __init__(self, collection):
        super(Selector, self).__init__()
        if not hasattr(collection, '__iter__'):
            collection = NamedCollection({collection.id: collection})
        self._collection = collection

    def __call__(self, *args, **kwargs):
        _filter = args[0] if len(args) == 1 else lambda x: True
        subset = NamedCollection()
        for item in self._collection.values():
            if _filter(item):
                subset[item.id] = item
        return Selector(subset)

    def __getitem__(self, item):
        if isinstance(item, int):
            i = item if item >= 0 else item + len(self)
            for v in self:
                if i == 0:
                    return v
                else:
                    i -= 1
            raise IndexError(item)
        try:
            return self._collection[item]
        except KeyError:
            return NamedCollection()

    def __contains__(self, item):
        return item in self._collection

    def __setitem__(self, key, value):
        raise NotImplementedError('Cannot modify read-only collection')

    def __repr__(self):
        return repr(list(self))

    def __iter__(self):
        return iter(self._collection.values())

    def values(self):
        return iter(self._collection.values())

    def items(self):
        return iter(self._collection.items())

    def __len__(self):
        return len(list(iter(self)))

    def __eq__(self, other):
        return set(iter(self)) == set(iter(other))


class PrefixedSelector(Selector):
    def __init__(self, prefix, collection):
        super(PrefixedSelector, self).__init__(collection)
        self.prefix = prefix

    def __getitem__(self, item):
        try:
            return super(PrefixedSelector, self).__getitem__('%s/%s' % (self.prefix, item))
        except KeyError:
            # Fallback to non-prefixed implementation
            return super(PrefixedSelector, self).__getitem__(item)

    def __contains__(self, item):
        return super(PrefixedSelector, self).__contains__('%s/%s' % (self.prefix, item)) or \
            super(PrefixedSelector, self).__contains__(item)

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __iter__(self):
        return iter(self._collection.values())

    def iteritems(self):
        for k, v in self._collection.items():
            if not k.startswith(self.prefix):
                yield '%s/%s' % (self.prefix, k), v
            else:
                yield k, v

    def itervalues(self):
        return iter(self._collection.values())


class NamedCollection(dict, AggregateOperations, Selectors):
    def __call__(self, functor=None, *args, **kwargs):
        """

        :param functor:
        :param args:
        :param kwargs:
        :return:
        :rtype: NamedCollectionView
        """
        if len(args) > 0:
            for i in range(len(args)):
                functor = functor & args[i]
        return NamedCollectionView(self, functor)

    def __getitem__(self, item):
        if isinstance(item, Expression):
            return NamedCollectionView(self, item)
        elif isinstance(item, tuple):
            return NamedCollectionView(self, reduce(AndExpression, item))
        else:
            return super(NamedCollection, self).__getitem__(item)


class NamedCollectionView(AggregateOperations, Selectors):
    def __init__(self, parent, functor):
        super(NamedCollectionView, self).__init__()
        self.parent = parent
        self.functor = functor

    def __iter__(self):
        parent, functor = self.parent, self.functor

        if functor is not None:
            def generator():
                for key, value in parent.items():
                    if functor(value):
                        yield key, value
                raise StopIteration
            return generator()
        else:
            return iter(self.parent.items())

    def iteritems(self):
        return iter(self)

    def filter(self, rule):
        return self if rule is None else NamedCollectionView(self, rule)

    def __repr__(self):
        return repr(dict(iter(self)))


class UnionSelector(AggregateOperations, Selectors):
    def __init__(self, collection, field):
        super(UnionSelector, self).__init__()
        self.collection = collection
        self.field = field

    def __iter__(self):
        for c in self.collection.values():
            if hasattr(c, self.field):
                for v in getattr(c, self.field):
                    yield v

    def itervalues(self):
        for c in self.collection:
            if hasattr(c, self.field):
                child = getattr(c, self.field)
                for v in child.values():
                    yield v

    def iteritems(self):
        for c in self.collection:
            if hasattr(c, self.field):
                child = getattr(c, self.field)
                for k, v in child.items():
                    yield k, v

    def __call__(self, *args, **kwargs):
        if len(args) == 1:
            functor = args[0]
        else:
            def functor(_):
                return True
        if len(kwargs) > 0:
            pass
        else:
            return Selector({v.id: v for v in iter(self) if functor(v)})

    def __getitem__(self, item):
        if isinstance(item, int):
            i = item if item >= 0 else item + len(self)
            for v in self:
                if i == 0:
                    return v
                else:
                    i -= 1
            raise IndexError(item)
        else:
            for v in self.collection.values():
                if hasattr(v, self.field):
                    haystack = getattr(v, self.field)
                    if item in haystack:
                        return haystack[item]
                    else:
                        # TODO(ewpatton): Make support ID paths, e.g. Screen1/Arrange1/Button1
                        for u in haystack.values():
                            if u.name == item:
                                return u
        raise KeyError(item)

    def __len__(self):
        return len(list(iter(self)))


def select(item):
    """
    :py:func:`~aiatools.selectors.select` is a convenience method for creating a selection out of a single item. This
    allows one to select a specific component for further exploration
    :param item:
    :return:
    """
    return Selector(item)



#!/usr/bin/env python
# -*- mode: python; -*-

"""
AIA Tools provides a set of Python classes to interface with application sources exported from
`App Inventor <http://appinventor.mit.edu/>`_.
"""

from aiatools.aia import *
from aiatools.algebra import *
from aiatools.attributes import *
from aiatools.block_types import *
from aiatools.component_types import *
from aiatools.selectors import *

__author__ = 'Evan W. Patton <ewpatton@mit.edu>'
__version__ = '0.4.1'


def _aia_main():
    """
    The main program for aiatools. See aiatool --help for more information.
    """
    raise NotImplemented()


if __name__ == '__main__':
    _aia_main()



#!/usr/bin/env python

from setuptools import setup, find_packages
from aiatools import __version__

with open('README.md') as f:
    long_description = f.read()

setup(name='aiatools',
      version=__version__,
      description='Tools for extracting information from App Inventor AIA files',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Evan W. Patton',
      author_email='ewpatton@mit.edu',
      url='https://github.com/mit-cml/aiatools',
      packages=find_packages(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Education',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Software Development',
          'Topic :: Utilities'
      ],
      license='GPLv3+',
      keywords='App Inventor AIA extraction analysis toolkit',
      entry_points={
          'console_scripts': [
              'aia = aiatools:aia_main'
          ]
      },
      install_requires=[
          'jprops>=2.0.2'
      ],
      package_data={
          'aiatools': ['simple_components.json']
      }
      )



# -*- mode: python; coding: utf-8; -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""
.. testsetup::

    from aiatools import *
    project = AIAFile('test_aias/LondonCholeraMap.aia')

The :py:mod:`aiatools.common` module defines the core data model that is used throughout the aiatools project. In most
cases, users will not construct these objects directly but rather through use of the :py:class:`~aiatools.aia.AIAFile`
class.
"""

import re
from .algebra import Atom
from functools import reduce
from enum import Enum


__author__ = 'Evan W. Patton <ewpatton@mit.edu>'


NAMESPACE = re.compile('\\{[^}]+}')


def _html(tag):
    """
    Prefixes the given ``tag`` with the XHTML prefix. Block files saved with newer versions of App Inventor will have
    this XML namespace prepended when read with the ETree framework.
    :param tag: An XML element tag
    :type tag: basestring
    :return: A new string prefixed with the XHTML prefix.
    :rtype: str or unicode
    """
    return '{http://www.w3.org/1999/xhtml}' + tag


# noinspection PyShadowingBuiltins
class Block(object):
    _CATEGORIES = {'color', 'component', 'controls', 'global', 'lexical', 'lists', 'local', 'logic', 'math',
                   'obfuscated', 'dictionaries', 'procedures', 'text', 'pair', 'helpers'}
    _CATEGORY_MAP = {
        'color': 'Colors',
        'component': 'Components',
        'controls': 'Controls',
        'global': 'Variables',
        'lexical': 'Variables',
        'lists': 'Lists',
        'local': 'Variables',
        'logic': 'Logic',
        'math': 'Math',
        'obfuscated': 'Text',
        'obsfucated': 'Text',
        'dictionaries': 'Dictionaries',
        'pair': 'Dictionaries',
        'procedures': 'Procedures',
        'text': 'Text',
        'helpers': 'Helpers'
    }
    _ID_COUNT = 0

    def __init__(self, id, type):
        self.id = id
        self.type = type
        parts = type.split('_')
        if parts[0] == 'text':
            self.category = 'Text'
        else:
            self.category = Block._CATEGORY_MAP[parts[0]]
        self.parent = None
        self.logical_parent = None
        self.output = None
        self.inputs = {}
        self.ordered_inputs = []
        self.fields = {}
        self.statements = {}
        self.values = {}
        self.next = None
        self.mutation = None
        self.x = None
        self.y = None
        self.inline = False
        self.comment = None
        self.disabled = False
        self.logically_disabled = False
        self.screen = None

    @classmethod
    def from_xml(cls, screen, xml, lang_ver, siblings=None, parent=None, connection_type=None):
        siblings = siblings if siblings is not None else []
        attributes = xml.attrib
        type = attributes['type']
        id = attributes['id'] if 'id' in attributes else None
        if id is None:
            id = Block._ID_COUNT
            Block._ID_COUNT += 1
        if type == 'procedures_do_then_return':
            type = 'controls_do_then_return'
        elif type == 'procedure_lexical_variable_get':
            type = 'lexical_variable_get'
        elif type == 'for_lexical_variable_get':
            type = 'lexical_variable_get'
        type_parts = type.split('_')
        extra_mutations = {}
        if type_parts[0] not in Block._CATEGORIES and (lang_ver is None or lang_ver < 17):
            # likely old-format blocks code with component names in block types
            component = screen.components(lambda x: x.name == type_parts[0])
            if len(component) > 0:
                extra_mutations['instance_name'] = type_parts[0]
                extra_mutations['component_type'] = component[0].type.name
                if type_parts[1] == 'setproperty':
                    # Old-style property setter
                    type = 'component_set_get'
                    extra_mutations['property_name'] = type_parts[1]
                    extra_mutations['set_or_get'] = 'set'
                elif type_parts[1] == 'getproperty':
                    # Old-style property getter
                    type = 'component_set_get'
                    extra_mutations['property_name'] = type_parts[1]
                    extra_mutations['set_or_get'] = 'get'
                elif len(xml) >= 2 and xml[1].tag == 'statement' and 'name' in xml[1].attrib and \
                        xml[1].attrib['name'] == 'DO':
                    # Old-style event handler
                    type = 'component_event'
                    extra_mutations['event_name'] = type_parts[1]
                else:
                    # Old-style method call
                    type = 'component_method'
                    extra_mutations['method_name'] = type_parts[1]
            elif type_parts[0] == 'obsufcated':
                type_parts = ['obfuscated', 'text']
                type = '_'.join(type_parts)
            else:
                raise RuntimeError('Unknown block type: %s' % type_parts[0])
        block = Block(id, type)
        screen._blocks[id] = block
        block.screen = screen
        block.parent = parent
        block.logical_parent = parent
        siblings.append(block)
        if 'x' in attributes and 'y' in attributes:  # Top level block
            block.x, block.y = attributes['x'], attributes['y']
        if 'inline' in attributes:
            block.inline = attributes['inline'] == 'true'
        if 'disabled' in attributes and attributes['disabled'] == 'true':
            block.disabled = True
            block.logically_disabled = True
        if connection_type == 'value' or connection_type == 'statement':
            block.logically_disabled = block.disabled or parent.logically_disabled
        for child in xml:
            if child.tag == 'mutation' or child.tag == _html('mutation'):
                block.mutation = dict(child.attrib)
                block.mutation.update(extra_mutations)
                extra_mutations = {}
                if type.startswith('component_') and ('is_generic' not in block.mutation or
                                                      block.mutation['is_generic'] == 'false'):
                    block.component = screen.components[block.mutation['instance_name']]
                for grandchild in child:
                    tag = '.' + NAMESPACE.sub('', grandchild.tag)
                    if tag not in block.mutation:
                        block.mutation[tag] = []
                    block.mutation[tag].append(dict(grandchild.attrib))
            elif child.tag == 'comment' or child.tag == _html('comment'):
                block.comment = child.text
            elif child.tag in {'field', 'title', _html('field'), _html('title')}:
                block.fields[child.attrib['name']] = child.text
            elif child.tag == 'value' or child.tag == _html('value'):
                block.inputs[child.attrib['name']] = block.values[child.attrib['name']] = []
                child_block = Block.from_xml(screen, child[0], lang_ver, block.values[child.attrib['name']],
                                             parent=block, connection_type='value')
                child_block.output = block
                block.ordered_inputs.append(child_block)
            elif child.tag == 'statement' or child.tag == _html('statement'):
                block.inputs[child.attrib['name']] = block.statements[child.attrib['name']] = []
                child_block = Block.from_xml(screen, child[0], lang_ver, block.statements[child.attrib['name']],
                                             parent=block, connection_type='statement')
                block.ordered_inputs.append(child_block)
                for child_block in block.statements[child.attrib['name']]:
                    child_block.logical_parent = block
            elif child.tag == 'next' or child.tag == _html('next'):
                child_block = Block.from_xml(screen, child[0], lang_ver, siblings=siblings, parent=block,
                                             connection_type='next')
                block.next = child_block
        if len(extra_mutations) > 0:  # mutations were not consumed, so we're missing a <mutation> tag
            block.mutation = extra_mutations
        return block

    @property
    def return_type(self):
        if self.type == 'component_method':
            from aiatools import component_types
            type = getattr(component_types, self.mutation['component_type'])
            assert(isinstance(type, ComponentType))
            return type.methods[self.mutation['method_name']].return_type
        return None

    @property
    def kind(self):
        from aiatools import block_types
        type = getattr(block_types, self.type)
        if type.kind == BlockKind.MUTATION:
            if self.type == 'component_set_get':
                return BlockKind.VALUE if self.mutation['set_or_get'] == 'get' else BlockKind.STATEMENT
            elif self.type == 'component_method':
                return BlockKind.VALUE if self.return_type is not None else BlockKind.STATEMENT
            else:
                raise ValueError('Unknown type ' + self.type)
        else:
            return type.kind

    def children(self):
        return reduce(list.__add__, iter(self.values.values()), []) + \
               reduce(list.__add__, iter(self.statements.values()), [])

    def has_mutation(self, name):
        """
        Tests whether the block has a given mutation.

        :param str name: the name of the mutation to test
        :return: True if the mutation is present on the block, otherwise False
        :rtype: bool
        """
        return False if self.mutation is None else (name in self.mutation)

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.id, self.type)

    def __str__(self):
        description = {
            'id': self.id,
            'type': self.type,
            'disabled': self.disabled
        }
        if self.x is not None:
            description['x'] = self.x
            description['y'] = self.y
        if self.comment is not None:
            description['comment'] = self.comment
        if self.mutation is not None:
            description['mutation'] = self.mutation
        if len(self.fields) > 0:
            description['fields'] = self.fields
        if len(self.values) > 0:
            description['values'] = self.values
        if len(self.statements) > 0:
            description['statements'] = self.statements
        return str(description)

    def _get_is_generic(self):
        return self.mutation and 'is_generic' in self.mutation and self.mutation['is_generic'] == 'true'

    def __hash__(self):
        return hash(self.id) * 31 + hash(self.type)

    generic = property(_get_is_generic, doc="""
    True if the block is a generic component block (getter, setter, or method call), otherwise False.

    :type: bool
    """)
    # statements = property(_get_statements, doc="""
    # Access the statements of a block with continuation in an iterable fashion.
    #
    # :type: list[:class:`Block`]
    # """)


class BlockKind(Atom, Enum):
    DECLARATION = 1
    STATEMENT = 2
    VALUE = 3
    MUTATION = 4


class BlockType(Atom):
    def __init__(self, name, category, kind):
        """

        :param name:
        :type name: str
        :param category:
        :type category: BlockCategory
        :param kind:
        :type kind: BlockKind
        """
        self.name = name
        self.category = category
        self.kind = kind
        category.add_type(self)

    def __repr__(self):
        return 'BlockType(%r, %r, %r)' % (self.name, self.category, self.kind)

    def __call__(self, *args, **kwargs):
        return self.name


class BlockCategory(Atom):
    def __init__(self, name):
        self.name = name
        self.blocks = {}

    def __repr__(self):
        return 'aiatools.block_types.%s' % self.name

    def __call__(self, *args, **kwargs):
        return self.name

    def add_type(self, block_type):
        """

        :param block_type:
        :type block_type: BlockType
        """
        self.blocks[block_type.name] = block_type


class Method(Atom):
    # noinspection PyPep8Naming
    def __init__(self, name, description, deprecated, params, returnType=None, continuation=False):
        self.name = name
        self.description = description
        if isinstance(deprecated, str):
            self.deprecated = deprecated == 'true'
        else:
            self.deprecated = deprecated
        self.continuation = continuation == 'true'
        self.params = [param if isinstance(param, Parameter) else Parameter(**param) for param in params]
        self.return_type = returnType

    def __call__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        return 'Method(%s, %s, %s, %s)' % \
               (repr(self.name), repr(self.description), repr(self.deprecated), repr(self.params))


class Property(Atom):
    # noinspection PyPep8Naming
    def __init__(self, name, editorType=None, defaultValue=None, description=None, type=None, rw=None,
                 deprecated=False, editorArgs=None, alwaysSend=False, category=None, helper=None):
        self.name = name
        self.type = type
        self.editor_type = editorType
        self.default_value = defaultValue
        self.category = category
        self.description = description
        self.helper = helper
        self.rw = rw
        if isinstance(deprecated, str):
            self.deprecated = deprecated == 'true'
        else:
            self.deprecated = deprecated
        self.editor_args = editorArgs
        self.always_send = alwaysSend

    def __call__(self, component, *args, **kwargs):
        return component.properties[self.name]

    def __repr__(self):
        return 'Property(%r, %r, %r, %r, %r, %r, %r)' % \
               (self.name, self.editor_type, self.default_value, self.description, self.type, self.rw, self.deprecated)


class Event(Atom):
    def __init__(self, name, description, deprecated, params):
        self.name = name
        self.description = description
        if isinstance(deprecated, str):
            self.deprecated = deprecated == 'true'
        else:
            self.deprecated = deprecated
        self.params = [param if isinstance(param, Parameter) else Parameter(**param) for param in params]

    def __call__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        return 'Event(%r, %r, %r, %r)' % (self.name, self.description, self.deprecated, self.params)


class Parameter(Atom):
    def __init__(self, name, type, helper=None):
        self.name = name
        self.type = type
        self.helper = helper

    def __call__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        return 'Parameter(%r, %r)' % (self.name, self.type)


class RecursiveIterator(object):
    def __init__(self, container, order='breadth', test=None, skip=None):
        self.stack = [container]
        self.order = order
        self.test = test
        self.skip = skip

    def __iter__(self):
        while len(self.stack) > 0:
            item = self.stack.pop(0)
            failed = self.test and not self.test(item)
            if self.skip and failed:
                continue
            if self.order == 'breadth':
                self.stack.extend(item.children())
            elif self.order == 'depth':
                self.stack = list(filter(lambda x: x, [child if child != item else None for child in item.children()])) + \
                             self.stack
            else:
                raise NotImplementedError(f'Recursive order {self.order} is unknown.')
            if not failed:
                yield item


class ComponentType(Atom):
    # noinspection PyShadowingBuiltins
    def __init__(self, name, methods=None, events=None, properties=None):
        self.name = name
        self.type = None
        self.external = False
        self.version = 1
        self.category_string = None
        self.help_string = None
        self.show_on_palette = True
        self.visible = False
        self.icon_name = None
        self.methods = None
        self.events = None
        self.properties = None
        if methods is not None:
            self.methods = {name: Method(**m) if isinstance(m, dict) else m for name, m in methods.items()}
            for name, method in self.methods.items():
                setattr(self, method.name, method)
        if events is not None:
            self.events = {name: Event(**e) if isinstance(e, dict) else e for name, e in events.items()}
            for name, event in self.events.items():
                setattr(self, event.name, event)
        if properties is not None:
            self.properties = {name: Property(**p) if isinstance(p, dict) else p for name, p in properties.items()}
            for name, property in self.properties.items():
                setattr(self, property.name, property)

    def __call__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        return 'aiatools.component_types.%s' % self.name

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, ComponentType):
            if self.type is None or other.type is None:
                return self.name == other.name
            else:
                return self.type == other.type
        else:
            return super(ComponentType, self).__eq__(other)


class Extension(ComponentType):
    def __init__(self, *args, **kwargs):
        super(Extension, self).__init__(*args, **kwargs)
        self.external = True
        self.project = None

    def __repr__(self):
        return '<extension %s>' % self.type


class Component(object):
    _DISALLOWED_KEYS = {'$Components', '$Name', '$Type', '$Version', 'Uuid'}
    TYPES = {}

    def __init__(self, parent, uuid, type, name, version, properties=None):
        """

        :param parent:
        :type parent: ComponentContainer|None
        :param uuid:
        :param type:
        :param name:
        :param version:
        :param properties:
        """
        self.id = uuid
        self.parent = parent
        self.uuid = uuid
        self.type = type
        self.name = name
        self.version = version
        self.properties = properties
        self.path = '%s/%s' % (parent.name, self.name) if parent is not None else self.name

    def __repr__(self):
        # return '%s(%r, %r, %r, %r, %r, %r)' % (self.__class__.__name__, self.parent, self.uuid, self.type, self.name,
        #                                        self.version, self.properties)
        return '%s(%r, %r)' % (self.type, self.uuid, self.name)

    def children(self):
        return []

    @classmethod
    def from_json(cls, parent, json_repr):
        typename = json_repr['$Type']
        type = Component.TYPES[typename] if typename in Component.TYPES else Extension(typename)
        properties = {k: v for k, v in json_repr.items() if k not in Component._DISALLOWED_KEYS}
        return cls(parent, json_repr['Uuid'], type, json_repr['$Name'], json_repr['$Version'], properties)


class FilterableDict(dict):
    def filter(self, rule):
        if rule is None:
            return self
        else:
            return FilterableDict({k: v for k, v in self.items() if rule(k, v)})
